import { Expose, Transform, plainToClass, classToPlain } from 'class-transformer';
import { getClassValidatorMessagesStorage, Validator, getMetadataStorage, ValidationTypes, validateSync } from 'class-validator-multi-lang';
import { BehaviorSubject, Observable, Subject, combineLatest } from 'rxjs';
import { FormControl, FormGroup, FormArray, FormBuilder } from '@angular/forms';
import 'reflect-metadata';
import stringify from 'fast-safe-stringify';
import { distinctUntilChanged, tap, take } from 'rxjs/operators';
import stringHash from 'string-hash';

const DEFAULT_CLASS_TRANSFORM_OPTIONS = {
    strategy: 'excludeAll',
};
const DEFAULT_CLASS_TRANSFORM_TO_PLAIN_OPTIONS = {
    strategy: 'exposeAll',
};
const DEFAULT_CLASS_VALIDATOR_OPTIONS = { validationError: { target: false }, always: true };

// for correct work expose with js object
// https://github.com/typestack/class-transformer/issues/365
function ExposeNested(options) {
    const exposeFn = Expose(options);
    const transformFn = (propertyKey) => Transform((params) => {
        return ((params.obj && Object.getOwnPropertyDescriptor(params.obj, propertyKey)
            ? params.obj[propertyKey]
            : options?.default) || params.value);
    });
    return (target, propertyKey) => {
        transformFn(propertyKey)(target, propertyKey);
        exposeFn(target, propertyKey);
    };
}

/**
 * This function returns the global object across Node and browsers.
 *
 * Note: `globalThis` is the standardized approach however it has been added to
 * Node.js in version 12. We need to include this snippet until Node 12 EOL.
 */
function getGlobal() {
    if (typeof globalThis !== 'undefined') {
        return globalThis;
    }
    if (typeof global !== 'undefined') {
        return global;
    }
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore: Cannot find name 'window'.
    if (typeof window !== 'undefined') {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore: Cannot find name 'window'.
        return window;
    }
    // eslint-disable-next-line @typescript-eslint/ban-ts-comment
    // @ts-ignore: Cannot find name 'self'.
    if (typeof self !== 'undefined') {
        // eslint-disable-next-line @typescript-eslint/ban-ts-comment
        // @ts-ignore: Cannot find name 'self'.
        return self;
    }
    throw Error('Can not work with the global object');
}

const CLASS_VALIDATOR_MESSAGES_STORAGE = 'CLASS_VALIDATOR_MESSAGES_STORAGE';
function getValidatorMessagesStorage() {
    const global = getGlobal();
    if (!global[CLASS_VALIDATOR_MESSAGES_STORAGE]) {
        global[CLASS_VALIDATOR_MESSAGES_STORAGE] = new BehaviorSubject(getClassValidatorMessagesStorage());
    }
    return global[CLASS_VALIDATOR_MESSAGES_STORAGE];
}
function updateValidatorMessagesStorage(messages) {
    getValidatorMessagesStorage().next(messages);
}

const CLASS_VALIDATOR_TITLES_STORAGE = 'CLASS_VALIDATOR_TITLES_STORAGE';
function getValidatorTitlesStorage() {
    const global = getGlobal();
    if (!global[CLASS_VALIDATOR_TITLES_STORAGE]) {
        global[CLASS_VALIDATOR_TITLES_STORAGE] = new BehaviorSubject({});
    }
    return global[CLASS_VALIDATOR_TITLES_STORAGE];
}
function updateValidatorTitlesStorage(titles) {
    getValidatorTitlesStorage().next(titles);
}

function isDynamicFormGroupConfig(options) {
    return (options &&
        (Object.getOwnPropertyDescriptor(options, 'classValidatorOptions') ||
            Object.getOwnPropertyDescriptor(options, 'classTransformOptions') ||
            Object.getOwnPropertyDescriptor(options, 'classTransformToPlainOptions') ||
            Object.getOwnPropertyDescriptor(options, 'validateAllFormFields')));
}
function isLegacyOrOpts(options) {
    return options && (!!options['validator'] || !!options['asyncValidator']);
}
function isAbstractControlOptions(options) {
    return options && (!!options.validators || !!options.asyncValidators || !!options.updateOn);
}

function foreverInvalid(c) {
    return {
        foreverInvalid: {
            valid: false,
        },
    };
}
const FOREVER_INVALID_NAME = 'foreverInvalid';

class DynamicFormControl extends FormControl {
    constructor(name, fieldDefinition) {
        super(fieldDefinition.data instanceof Observable ? null : fieldDefinition.data, fieldDefinition.validationFunctions
            .filter((func) => func.type === 'sync')
            .map((func) => func.validator), fieldDefinition.validationFunctions
            .filter((func) => func.type === 'async')
            .map((func) => func.validator));
        this.controlName = name;
        this.validationDefinitions = fieldDefinition.validationDefinitions;
    }
}

const cloneDeep$2 = require('lodash.clonedeep');
const mergeWith = require('lodash.mergewith');
function transformValidationErrors(errors) {
    const customErrors = {};
    errors.forEach((error) => {
        if (error && error.constraints !== undefined) {
            Object.keys(error.constraints).forEach((key) => {
                if (!customErrors[error.property]) {
                    customErrors[error.property] = [];
                }
                if (error.constraints && customErrors[error.property].indexOf(error.constraints[key]) === -1) {
                    customErrors[error.property].push(error.constraints[key]);
                }
            });
        }
        if (error.children !== undefined && error.children.length) {
            customErrors[error.property] = transformValidationErrors(error.children);
        }
    });
    return customErrors;
}
function mergeErrors(externalErrors, validationErrors) {
    const clonedExternalErrors = cloneDeep$2(externalErrors);
    return mergeWith(clonedExternalErrors, validationErrors, (objValue, srcValue) => {
        if (canMerge()) {
            return objValue.concat(srcValue);
        }
        function canMerge() {
            return (Array.isArray(objValue) && Array.isArray(srcValue) && !objValue.find((objItem) => srcValue.includes(objItem)));
        }
    });
}

function getOrSetEmptyObject(factoryModel) {
    if (!factoryModel) {
        return undefined;
    }
    if (!factoryModel.emptyObject) {
        factoryModel.emptyObject = new factoryModel();
    }
    return factoryModel.emptyObject;
}

function hasToJSON(object) {
    return object ? typeof object.toJSON === 'function' : false;
}

const cloneDeep$1 = require('lodash.clonedeep');
const validator = new Validator();
class DynamicFormGroup extends FormGroup {
    constructor(factoryModel, fields, _validatorOrOpts, _asyncValidator, options) {
        super({}, _validatorOrOpts, _asyncValidator);
        this.factoryModel = factoryModel;
        this.fields = fields;
        this._validatorOrOpts = _validatorOrOpts;
        this._asyncValidator = _asyncValidator;
        this.options = options;
        this.nativeValidateErrors = new BehaviorSubject({});
        this.customValidateErrors = new BehaviorSubject({});
        this.objectChange = new Subject();
        /*
        const classValidators = DynamicFormGroup.getClassValidators<TModel>(
          this.factoryModel,
          this.fields,
          this.defaultValidatorOptions
        );
        const formGroup = this._fb.group(
          classValidators
        );
        Object.keys(formGroup.controls).forEach(key => {
          this.addControl(key, formGroup.controls[key]);
        });
        this.valueChanges.subscribe(data => {
          this.validate(
            undefined,
            this.defaultValidatorOptions
          );
        });*/
        if (options?.classTransformOptions) {
            this._classTransformOptions = options.classTransformOptions;
        }
        if (options?.classTransformToPlainOptions) {
            this._classTransformToPlainOptions = options.classTransformToPlainOptions;
        }
        if (options?.classValidatorOptions) {
            this._validatorOptions = options.classValidatorOptions;
        }
        if (options?.validateAllFormFields !== undefined) {
            this._validateAllFormFields = options.validateAllFormFields;
        }
        this.formFields = this.onlyFields(fields);
    }
    subscribeToValueChanges(externalErrors, validatorOptions) {
        if (!this.valueChangesSubscription) {
            this.valueChangesSubscription = this.valueChanges
                .pipe(distinctUntilChanged(), tap(() => this.validate(externalErrors, validatorOptions)))
                .subscribe({
                error: (err) => {
                    throw err;
                },
            });
        }
    }
    // Getters & Setters
    set externalErrors(externalErrors) {
        this._externalErrors = externalErrors;
        this.validate();
    }
    get externalErrors() {
        return this._externalErrors;
    }
    set classTransformOptions(classTransformOptions) {
        this._classTransformOptions = classTransformOptions;
        this.validate();
    }
    get classTransformOptions() {
        return this._classTransformOptions;
    }
    set classTransformToPlainOptions(classTransformToPlainOptions) {
        this._classTransformToPlainOptions = classTransformToPlainOptions;
        this.validate();
    }
    get classTransformToPlainOptions() {
        return this._classTransformToPlainOptions;
    }
    set validatorOptions(validatorOptions) {
        this._validatorOptions = validatorOptions;
        this.validate();
    }
    get validatorOptions() {
        return this._validatorOptions;
    }
    set object(object) {
        this.setObject(object);
    }
    get object() {
        return this.getObject();
    }
    set json(jsonData) {
        this.setJSON(jsonData);
    }
    get json() {
        return this.getJSON();
    }
    validateStream(externalErrors, validatorOptions) {
        const argumentExternalErrors = externalErrors ? cloneDeep$1(externalErrors) : externalErrors;
        const argumentValidatorOptions = validatorOptions
            ? cloneDeep$1(validatorOptions)
            : validatorOptions;
        return combineLatest([getValidatorMessagesStorage(), getValidatorTitlesStorage()]).pipe(tap(([messages, titles]) => {
            validatorOptions = cloneDeep$1(this._validatorOptions || argumentValidatorOptions);
            externalErrors = cloneDeep$1(this._externalErrors || argumentExternalErrors);
            if (!validatorOptions) {
                validatorOptions = {};
            }
            if (!validatorOptions.messages) {
                validatorOptions.messages = {};
            }
            if (!validatorOptions.titles) {
                validatorOptions.titles = {};
            }
            if (messages) {
                validatorOptions.messages = { ...cloneDeep$1(messages), ...validatorOptions.messages };
            }
            if (titles) {
                validatorOptions.titles = { ...cloneDeep$1(titles), ...validatorOptions.titles };
            }
            const dataToValidate = this.object;
            const validationErrors = getValidateErrors(this, dataToValidate, validatorOptions);
            if (!externalErrors) {
                externalErrors = {};
            }
            const allErrors = mergeErrors(externalErrors, validationErrors);
            const allErrorsKeys = Object.keys(allErrors);
            this.markAsInvalidForExternalErrors(externalErrors);
            this.setCustomErrors(allErrors);
            // todo: refactor, invalidate form if exists any allErrors
            let usedForeverInvalid = false;
            if (!allErrorsKeys.find((key) => key !== FOREVER_INVALID_NAME) && this.get(FOREVER_INVALID_NAME)) {
                this.removeControl(FOREVER_INVALID_NAME);
                usedForeverInvalid = true;
            }
            if (this.valid && allErrorsKeys.length > 0 && !this.get(FOREVER_INVALID_NAME)) {
                this.addControl(FOREVER_INVALID_NAME, new FormControl('', [foreverInvalid]));
                usedForeverInvalid = true;
            }
            if (usedForeverInvalid) {
                this.updateValueAndValidity({
                    onlySelf: true,
                    emitEvent: false,
                });
            }
            if (this._validateAllFormFields) {
                this.validateAllFormFields();
            }
        }));
    }
    validateAsync(externalErrors, validatorOptions) {
        return this.validateStream(externalErrors, validatorOptions).pipe(take(1)).toPromise();
    }
    // Public API
    validate(externalErrors, validatorOptions) {
        if (this._validateSubscription) {
            this._validateSubscription.unsubscribe();
            this._validateSubscription = undefined;
        }
        this._validateSubscription = this.validateStream(externalErrors, validatorOptions).subscribe({
            error: (err) => {
                throw err;
            },
        });
    }
    setCustomErrors(allErrors, force = false) {
        if (force || stringify(allErrors) !== stringify(this.customValidateErrors.value)) {
            this.formErrors = allErrors;
            this.customValidateErrors.next(this.formErrors);
        }
        const nativeValidateErrors = this.collectErrors(this);
        if (force || stringify(nativeValidateErrors) !== stringify(this.nativeValidateErrors.value)) {
            this.nativeValidateErrors.next(this.collectErrors(this));
        }
    }
    collectErrors(control, isRoot = true) {
        if (control.controls) {
            return {
                ...(isRoot ? this.errors : {}),
                ...Object.entries(control.controls).reduce((acc, [key, childControl]) => {
                    const childErrors = this.collectErrors(childControl, false);
                    if (childErrors && Object.keys(childErrors).length > 0) {
                        acc = {
                            ...acc,
                            [key]: {
                                ...(acc && acc[key] ? acc[key] : {}),
                                ...childErrors,
                            },
                        };
                    }
                    return acc;
                }, {}),
            };
        }
        else {
            return control.errors;
        }
    }
    validateAllFormFields() {
        let control;
        let formArrayControlsLength;
        Object.keys(this.controls).forEach((field) => {
            control = this.get(field);
            // Control
            if (control instanceof FormControl) {
                control.markAsTouched({ onlySelf: true });
            }
            // Group: recursive
            else if (control instanceof DynamicFormGroup) {
                control.validateAllFormFields();
            }
            // Array
            else if (control instanceof FormArray) {
                formArrayControlsLength = control.controls.length;
                for (let i = 0; i < formArrayControlsLength; i++) {
                    // Control in Array
                    if (control.controls[i] instanceof FormControl) {
                        control.controls[i].markAsTouched({ onlySelf: true });
                    }
                    // Group in Array: recursive
                    else if (control.controls[i] instanceof DynamicFormGroup) {
                        control.controls[i].validateAllFormFields();
                    }
                }
            }
        });
        formArrayControlsLength = null;
        control = null;
    }
    resetValidateAllFormFields() {
        this.markAsInvalidForExternalErrors({});
        let control;
        Object.keys(this.controls).forEach((field) => {
            control = this.get(field);
            // Control
            if (control instanceof FormControl) {
                control.setErrors(null, { emitEvent: false });
                control.markAsUntouched({ onlySelf: true });
                control.markAsPristine({ onlySelf: true });
            }
            // Group: recursive
            else if (control instanceof DynamicFormGroup) {
                control.resetValidateAllFormFields();
            }
            // Array
            else if (control instanceof FormArray) {
                const formArrayControlsLength = control.controls.length;
                for (let i = 0; i < formArrayControlsLength; i++) {
                    // Control in Array
                    if (control.controls[i] instanceof FormControl) {
                        control.controls[i].setErrors(null, { emitEvent: false });
                        control.controls[i].markAsUntouched({ onlySelf: true });
                        control.controls[i].markAsPristine({ onlySelf: true });
                    }
                    // Group in Array: recursive
                    else if (control.controls[i] instanceof DynamicFormGroup) {
                        control.controls[i].resetValidateAllFormFields();
                    }
                }
            }
        });
        control = null;
        this.setCustomErrors({});
    }
    classToClass(object) {
        if (hasToJSON(object)) {
            return new object.constructor(object.toJSON());
        }
        return cloneDeep$1(object);
    }
    plainToClass(cls, plain) {
        if (hasToJSON(getOrSetEmptyObject(cls))) {
            return new cls(plain);
        }
        if (Object.keys(plain || {}).length === 0) {
            return this.classToClass(getOrSetEmptyObject(cls));
        }
        if (plain instanceof cls) {
            return this.classToClass(plain);
        }
        if (stringify(plain) === stringify(getOrSetEmptyObject(cls))) {
            return this.classToClass(getOrSetEmptyObject(cls));
        }
        return plainToClass(cls, plain, { ...this._classTransformOptions });
    }
    classToPlain(object) {
        if (hasToJSON(object)) {
            return object.toJSON();
        }
        return classToPlain(object, { ...this._classTransformToPlainOptions });
    }
    /**
     * @deprecated please use sync method setExternalErrors(errors)
     */
    setExternalErrorsAsync(externalErrors) {
        this.setExternalErrors(externalErrors);
        return Promise.resolve();
    }
    setExternalErrors(externalErrors) {
        this._externalErrors = externalErrors;
        this.validate();
    }
    getExternalErrors() {
        return this._externalErrors;
    }
    /**
     * @deprecated please use sync method setExternalErrors({})
     */
    clearExternalErrorsAsync() {
        this.setExternalErrors({});
        return Promise.resolve();
    }
    clearExternalErrors() {
        this.setExternalErrors({});
    }
    setValidatorOptions(validatorOptions) {
        this._validatorOptions = validatorOptions;
        this.validate();
    }
    getValidatorOptions() {
        return this._validatorOptions;
    }
    // Helpers
    onlyFields(fields) {
        const newFields = {};
        if (fields !== undefined) {
            Object.keys(fields).forEach((key) => {
                if (fields[key] instanceof DynamicFormGroup) {
                    // Group: recursive
                    newFields[key] = this.onlyFields(fields[key].formFields);
                }
                else {
                    // Array
                    if (fields[key] instanceof FormArray) {
                        if (fields[key].controls[0] instanceof DynamicFormGroup) {
                            // Group within Array: recursive
                            newFields[key] = this.onlyFields(fields[key].controls[0].formFields);
                        }
                        else {
                            // Control within Array
                            newFields[key] = fields[key].controls[0].value;
                        }
                    }
                    else {
                        // Handle Control
                        newFields[key] = fields[key];
                    }
                }
            });
        }
        return newFields;
    }
    markAsInvalidForExternalErrors(errors, controls) {
        const currentControls = controls ? controls : this.controls;
        Object.keys(currentControls).forEach((field) => {
            const control = currentControls[field];
            // Control
            if (control instanceof FormControl) {
                if (errors && errors[field]) {
                    control.setErrors({ externalError: true });
                }
                else {
                    if (control.errors && control.errors.externalError === true) {
                        control.setErrors(null);
                    }
                }
            }
            // Group
            else if (control instanceof DynamicFormGroup) {
                control.markAsInvalidForExternalErrors(errors && errors[field] ? errors[field] : {});
            }
            // Array
            else if (control instanceof FormArray) {
                const formArrayControlsLength = control.length;
                for (let i = 0; i < formArrayControlsLength; i++) {
                    // Control in Array
                    if (control[i] instanceof FormControl) {
                        if (errors && errors[i] && errors[i][field]) {
                            control[i].setErrors({ externalError: true });
                        }
                        else if (control[i].errors && control[i].errors.externalError === true) {
                            control[i].setErrors(null);
                        }
                    }
                    // Group in Array
                    else if (control[i] instanceof DynamicFormGroup) {
                        control[i].markAsInvalidForExternalErrors(errors && errors[i] && errors[i][field] ? errors[i][field] : {});
                    }
                }
            }
        });
    }
    /**
     * Recursively gets all values from the form controls and all sub form group and array controls and returns it as
     * an object
     */
    getObject() {
        // Initialize the shape of the response
        const object = this.classToClass(this._object || getOrSetEmptyObject(this.factoryModel) || undefined);
        if (object !== undefined) {
            // Recursively get the value of all fields
            Object.keys(this.controls)
                .filter((name) => name !== FOREVER_INVALID_NAME)
                .forEach((key) => {
                // Handle Group
                if (this.controls[key] instanceof DynamicFormGroup) {
                    object[key] = this.controls[key].object;
                }
                // Handle Form Array
                else if (this.controls[key] instanceof FormArray) {
                    // Initialize value
                    object[key] = [];
                    const formArrayControlsLength = this.controls[key].controls.length;
                    let value;
                    for (let i = 0; i < formArrayControlsLength; i++) {
                        if (this.controls[key].controls[i] instanceof DynamicFormGroup) {
                            // Recursively get object group
                            value = this.controls[key].controls[i].object;
                        }
                        else {
                            value = this.controls[key].controls[i].value;
                        }
                        if (value && Object.keys(value).length > 0) {
                            object[key].push(value);
                        }
                    }
                    value = null;
                }
                // Handle Control
                else {
                    if (this.controls[key]) {
                        object[key] = this.controls[key].value;
                    }
                }
            });
        }
        return this.factoryModel ? this.plainToClass(this.factoryModel, object) : object;
        // todo: try change to return this.classToClass(object);
    }
    /**
     * Sets the value of every control on the form and recursively sets the values of the controls
     * on all sub form groups
     *
     * @param object the data to assign to all controls of the form group and sub groups
     */
    setObject(object) {
        // console.time(String(object));
        if (object instanceof this.factoryModel) {
            this._object = this.classToClass(object); // Ensure correct type
        }
        else {
            this._object = this.plainToClass(this.factoryModel, object); // Convert to Model type
            /*
            todo: remove later if speed is normal
            let objectWithOnlyFormFields: any = {};
            // for speed up on work with short version of model
            Object.keys(this.formFields).forEach((key) => (objectWithOnlyFormFields[key] = object ? object[key] : undefined));
            this._object = this.plainToClass(this.factoryModel, objectWithOnlyFormFields); // Convert to Model type
            objectWithOnlyFormFields = null;
            */
        }
        let newObject;
        let newValue;
        // Recursively set the value of all fields
        Object.keys(this.controls)
            .filter((name) => name !== FOREVER_INVALID_NAME)
            .forEach((key) => {
            // Handle Group
            if (this.controls[key] instanceof DynamicFormGroup) {
                if (this.controls[key]._object !== this._object[key]) {
                    this.controls[key].setObject(this._object ? this._object[key] : {});
                }
            }
            // Handle FormArray
            else if (this.controls[key] instanceof FormArray) {
                let objectArray = this._object ? this._object[key] : [];
                let formArray = this.controls[key];
                let isFormGroup = formArray.controls[0] instanceof DynamicFormGroup;
                let firstFormGroup = formArray.controls[0];
                let formControl = formArray.controls[0];
                // Clear FormArray while retaining the reference
                let objectArrayLength = objectArray.length;
                let dynamicFormGroup;
                let controlValue;
                let newFormControl;
                let classValidators;
                let formGroup;
                let formArrayLength = formArray.length;
                for (let i = 0; i < formArrayLength; i++) {
                    if (formArray.controls[i] instanceof DynamicFormGroup &&
                        formArray.controls[i].valueChangesSubscription) {
                        formArray.controls[i].valueChangesSubscription.unsubscribe();
                    }
                }
                while (formArrayLength !== 0) {
                    formArray.removeAt(0);
                    formArrayLength--;
                }
                formArrayLength = null;
                for (let i = 0; i < objectArrayLength; i++) {
                    if (isFormGroup) {
                        // Create FormGroup
                        dynamicFormGroup = this.dynamicFormBuilder.factoryDynamicFormGroup(firstFormGroup.factoryModel, firstFormGroup.formFields);
                        dynamicFormGroup.setParent(formArray);
                        formGroup = this.originalFormBuilder.group(getClassValidators(firstFormGroup.factoryModel, firstFormGroup.formFields));
                        // Add all controls to the form group
                        Object.keys(formGroup.controls).forEach((ctrlKey) => {
                            dynamicFormGroup.addControl(ctrlKey, formGroup.controls[ctrlKey]);
                        });
                        formArray.controls.push(dynamicFormGroup);
                    }
                    else {
                        // Create control
                        controlValue = this._object && objectArray && objectArray[i] ? objectArray[i] : undefined;
                        newFormControl = new FormControl(controlValue, formControl ? formControl.validator : undefined);
                        newFormControl.setParent(formArray);
                        // Add the control to the FormArray
                        formArray.controls.push(newFormControl);
                    }
                }
                let arrayLength = formArray.length;
                for (let i = 0; i < arrayLength; i++) {
                    if (isFormGroup) {
                        // Add a value change listener to the group. on change, validate
                        if (formArray.controls[i] instanceof DynamicFormGroup) {
                            // Recusrively set the object value
                            formArray.controls[i].setObject(this._object && objectArray && objectArray[i] ? objectArray[i] : {});
                        }
                    }
                }
                for (let i = 0; i < arrayLength; i++) {
                    if (isFormGroup) {
                        // Add a value change listener to the group. on change, validate
                        if (formArray.controls[i] instanceof DynamicFormGroup) {
                            formArray.controls[i].subscribeToValueChanges(undefined, this._validatorOptions);
                        }
                    }
                }
                classValidators = null;
                formGroup = null;
                controlValue = null;
                newFormControl = null;
                dynamicFormGroup = null;
                objectArrayLength = null;
                arrayLength = null;
                objectArray = null;
                formArray = null;
                isFormGroup = null;
                firstFormGroup = null;
                formControl = null;
            }
            // Handle Control
            else {
                newObject = this._object ? this._object[key] : [];
                newValue = this._object && newObject ? newObject : undefined;
                if (this.controls[key] && this.controls[key].value !== newValue) {
                    // very slow operations :(
                    this.controls[key].setValue(newValue, { emitEvent: false, onlySelf: true });
                }
            }
        });
        newValue = null;
        newObject = null;
        this.updateValueAndValidity();
        this.objectChange.next(this._object);
        if (this._validateAllFormFields) {
            this.validateAllFormFields();
        }
        // console.timeEnd(String(object));
    }
    getJSON() {
        return this.classToPlain(this.getObject());
    }
    setJSON(jsonData) {
        this.setObject(this.plainToClass(this.factoryModel, jsonData));
    }
}
function getClassValidators(factoryModel, fields, validatorOptions) {
    const groups = validatorOptions ? validatorOptions.groups : undefined;
    const strictGroups = (validatorOptions && validatorOptions.strictGroups) || false;
    const always = (validatorOptions && validatorOptions.always) || false;
    // console.time(String(factoryModel));
    // Get the validation rules from the object decorators
    let allValidationMetadatas = getMetadataStorage().getTargetValidationMetadatas(factoryModel, '', always, strictGroups);
    // Get the validation rules for the validation group: https://github.com/typestack/class-validator-multi-lang#validation-groups
    let validationGroupMetadatas = getMetadataStorage().getTargetValidationMetadatas(factoryModel, '', always, strictGroups, groups);
    const formGroupFields = {};
    // Loop through all fields in the form definition
    if (fields) {
        let conditionalValidations;
        let allNestedValidations;
        let nestedGroupValidations;
        let fieldDefinition;
        let dynamicValidate;
        let customValidation;
        let objectToValidate;
        let nestedValidate;
        Object.keys(fields)
            .filter((key) => key.indexOf('__') !== 0)
            .forEach((fieldName) => {
            // Conditional Validation for the field
            conditionalValidations = [];
            validationGroupMetadatas.forEach((validationMetadata) => {
                if (isPropertyValidatorOfType(validationMetadata, fieldName, ValidationKeys.conditional.type)) {
                    conditionalValidations.push(validationMetadata);
                }
            });
            // All Nested Validation for the field
            allNestedValidations = [];
            allValidationMetadatas.forEach((validationMetadata) => {
                if (isPropertyValidatorOfType(validationMetadata, fieldName, ValidationKeys.nested.type)) {
                    allNestedValidations.push(validationMetadata);
                }
            });
            // Nested Validation for the field for the requested class-validator-multi-lang group
            nestedGroupValidations = [];
            validationGroupMetadatas.forEach((validationMetadata) => {
                if (isPropertyValidatorOfType(validationMetadata, fieldName, ValidationKeys.nested.type)) {
                    nestedGroupValidations.push(validationMetadata);
                }
            });
            fieldDefinition = {
                data: formGroupFields[fieldName] || undefined,
                validationFunctions: [],
                validationDefinitions: [],
            };
            if (fieldDefinition.data === undefined) {
                fieldDefinition.data = fields[fieldName];
            }
            // TRY LINK EXISTS NATIVE VALIDATIONS, UNSTABLE !!!
            if (Array.isArray(fieldDefinition.data) &&
                fieldDefinition.data.length > 1 &&
                fieldDefinition.data.find((validationFunction, index) => index > 0 && typeof validationFunction === 'function')) {
                fieldDefinition.data
                    .filter((validationFunction, index) => index > 0 && typeof validationFunction === 'function')
                    .forEach((validationFunction) => fieldDefinition.validationFunctions.push({ type: 'sync', validator: validationFunction }));
                fieldDefinition.data = fieldDefinition.data[0];
            }
            validationGroupMetadatas.forEach((validationMetadata) => {
                if (validationMetadata.propertyName === fieldName &&
                    validationMetadata.type !== ValidationKeys.conditional.type) {
                    // Add all validation to the field except the @NestedValidation definition as
                    // being part of the form would imply it is validated if any other rules are present
                    if (validationMetadata.type !== ValidationKeys.nested.type) {
                        fieldDefinition.validationDefinitions.push(validationMetadata);
                    }
                    Object.keys(ValidationTypes).forEach((typeKey) => {
                        if (ValidationTypes.hasOwnProperty(typeKey)) {
                            // Handle Nested Validation
                            if (checkWithAllNestedValidations(allNestedValidations, nestedGroupValidations, fieldName)) {
                                if (isNestedValidate(validationMetadata, typeKey)) {
                                    objectToValidate =
                                        fields[fieldName] instanceof DynamicFormGroup ? fields[fieldName].object : undefined;
                                    nestedValidate = createNestedValidate(fieldName, objectToValidate, validationMetadata);
                                    setFieldData(fieldName, fieldDefinition, nestedValidate);
                                    objectToValidate = null;
                                    nestedValidate = null;
                                }
                            }
                            // Handle Custom Validation
                            if (isCustomValidate(validationMetadata, typeKey)) {
                                customValidation = createCustomValidation(fieldName, validationMetadata);
                                setFieldData(fieldName, fieldDefinition, customValidation);
                                customValidation = null;
                            }
                            // Handle remaining validation
                            if (isDynamicValidate(validationMetadata, typeKey)) {
                                dynamicValidate = createDynamicValidate(fieldName, validationMetadata, conditionalValidations);
                                setFieldData(fieldName, fieldDefinition, dynamicValidate);
                                dynamicValidate = null;
                            }
                        }
                    });
                }
            });
            // Convert to a structure, angular understands
            if (fieldDefinition.data instanceof DynamicFormGroup || fieldDefinition.data instanceof FormArray) {
                formGroupFields[fieldName] = fieldDefinition.data;
            }
            else {
                formGroupFields[fieldName] = new DynamicFormControl(fieldName, fieldDefinition);
            }
        });
        dynamicValidate = null;
        customValidation = null;
        objectToValidate = null;
        nestedValidate = null;
        conditionalValidations = null;
        allNestedValidations = null;
        nestedGroupValidations = null;
        fieldDefinition = null;
    }
    allValidationMetadatas = null;
    validationGroupMetadatas = null;
    // console.timeEnd(String(factoryModel));
    return formGroupFields;
    // ******************************************************************************************
    // Local Helper functions to help make the main code more readable
    //
    function createNestedValidate(fieldName, objectToValidate, validationMetadata) {
        return {
            type: 'sync',
            validator: function (control) {
                let parent = control.parent;
                let validateErrors = (parent &&
                    parent.value &&
                    getValidateErrors(parent, objectToValidate !== undefined ? objectToValidate : control.value, validatorOptions)) ||
                    {};
                let validateState = Object.keys(validateErrors || {}).length === 0;
                const result = getIsValidResult(validateState, validationMetadata, 'nestedValidate');
                parent = null;
                validateErrors = null;
                validateState = null;
                return result;
            },
        };
    }
    function createDynamicValidate(fieldName, validationMetadata, conditionalValidations) {
        return {
            type: 'sync',
            validator: function (control) {
                if (!control) {
                    return null;
                }
                let validateErrors = setObjectValueAndGetValidationErrors(fieldName, control, validatorOptions);
                let validateState = getAllErrors(validateErrors, fieldName).length === 0;
                const result = getIsValidResult(validateState, validationMetadata, 'dynamicValidate');
                validateErrors = null;
                validateState = null;
                return result;
            },
        };
    }
    function createCustomValidation(fieldName, validationMetadata) {
        return {
            type: 'sync',
            validator: function (control) {
                let validateErrors = setObjectValueAndGetValidationErrors(fieldName, control, validatorOptions);
                let validateState = validateErrors && validateErrors[fieldName] ? false : true;
                const result = getIsValidResult(validateState, validationMetadata, 'customValidation');
                validateErrors = null;
                validateState = null;
                return result;
            },
        };
    }
    function checkWithAllNestedValidations(allNestedValidations, nestedValidations, key) {
        const allNestedValidationsLength = allNestedValidations.length;
        const nestedValidationsLength = nestedValidations.length;
        return (allNestedValidationsLength === nestedValidationsLength ||
            (((fields && fields[key] instanceof DynamicFormGroup) || (fields && fields[key] instanceof FormArray)) &&
                allNestedValidationsLength > 0 &&
                nestedValidationsLength === 0));
    }
    function isDynamicValidate(validationMetadata, typeKey) {
        return validationMetadata.type === ValidationTypes[typeKey] && validator[validationMetadata.type] !== undefined;
    }
    /**
     * marked with @Validate(...)
     * https://github.com/typestack/class-validator-multi-lang#custom-validation-classes
     */
    function isCustomValidate(validationMetadata, typeKey) {
        return (isNotPropertyValidation(validationMetadata, typeKey) &&
            validationMetadata.type === ValidationKeys.custom.type &&
            typeKey === ValidationKeys.custom.typeKey);
    }
    /**
     * marked with @ValidateNested()
     * https://github.com/typestack/class-validator-multi-lang#validating-nested-objects
     */
    function isNestedValidate(validationMetadata, typeKey) {
        return (isNotPropertyValidation(validationMetadata, typeKey) &&
            validationMetadata.type === ValidationKeys.nested.type &&
            typeKey === ValidationKeys.nested.typeKey);
    }
    function isNotPropertyValidation(validationMetadata, typeKey) {
        return validationMetadata.type === ValidationTypes[typeKey] && validator[validationMetadata.type] === undefined;
    }
    function setFieldData(fieldName, fieldDefinition, validationFunction) {
        /* todo: maybe not need, if enable this code, experemental mode not work
        if (fields[fieldName] instanceof DynamicFormGroup) {
          fields[fieldName].object = fields[fieldName].fields;
        }*/
        // Fill field data if empty
        if (fieldDefinition.data === undefined) {
            fieldDefinition.data = fields && fields[fieldName];
        }
        fieldDefinition.validationFunctions.push(validationFunction);
    }
    function getAllErrors(validateErrors, fieldName) {
        return Object.keys(validateErrors).filter((key) => 
        // Check for nested/child errors
        validateErrors[key][fieldName] || validateErrors[fieldName]);
    }
}
// ***************************************************************
// Global Helper functions
//
function isPropertyValidatorOfType(validationMetadata, fieldName, validationMetadataType) {
    return validationMetadata.propertyName === fieldName && validationMetadata.type === validationMetadataType;
}
function setObjectValueAndGetValidationErrors(fieldName, control, validatorOptions) {
    let parent = control.parent instanceof DynamicFormGroup
        ? control.parent
        : control.parent
            ? control.parent
            : null;
    let object = control.parent instanceof DynamicFormGroup
        ? control.parent.object
        : control.parent
            ? control.parent.value
            : {};
    if (object) {
        object[fieldName] = control.value;
    }
    const result = (parent && parent.value && getValidateErrors(parent, object, validatorOptions)) || {};
    parent = null;
    object = null;
    return result;
}
function getValidateErrors(control, dataToValidate, validatorOptions) {
    let validateKey = stringHash(stringify({ dataToValidate, validatorOptions }));
    let ctrl = control || {};
    let validatedErrors;
    if (ctrl.__prevValidateKey === validateKey && ctrl.__prevValidateErrors !== undefined) {
        // console.log('CACHED', validateKey);
        validatedErrors = ctrl.__prevValidateErrors;
    }
    else {
        // console.log('NO CACHED', validateKey);
        validatedErrors = transformValidationErrors(validateSync(dataToValidate, validatorOptions));
        ctrl.__prevValidateErrors = validatedErrors;
        ctrl.__prevValidateKey = validateKey;
    }
    validateKey = null;
    ctrl = null;
    return validatedErrors;
}
function getIsValidResult(isValid, validationMetadata, errorType) {
    return isValid
        ? null
        : {
            [errorType]: {
                valid: false,
                type: validationMetadata.type,
            },
        };
}
const ValidationKeys = {
    nested: {
        type: 'nestedValidation',
        typeKey: 'NESTED_VALIDATION',
    },
    conditional: {
        type: 'conditionalValidation',
    },
    custom: {
        type: 'customValidation',
        typeKey: 'CUSTOM_VALIDATION',
    },
};

const cloneDeep = require('lodash.clonedeep');
class DynamicFormBuilder extends FormBuilder {
    constructor(options) {
        super();
        this.options = options;
        // need for createEmptyObject
        this.emptyDynamicFormGroup = this.factoryDynamicFormGroup(Object);
    }
    // ******************
    // Public API
    group(factoryModel, controlsConfig, options) {
        // console.time(factoryModel.toString());
        if (!controlsConfig && !options) {
            options = {};
        }
        // Process the group with the controlsConfig passed into extra instead. (What does this accomplish?)
        if (controlsConfig &&
            (isAbstractControlOptions(controlsConfig) ||
                isLegacyOrOpts(controlsConfig) ||
                isDynamicFormGroupConfig(controlsConfig)) &&
            Object.keys(options || {}).length === 0) {
            return this.group(factoryModel, undefined, controlsConfig);
        }
        let extra = cloneDeep(options);
        let validators = null;
        let asyncValidators = null;
        let updateOn;
        if (extra != null) {
            if (isAbstractControlOptions(extra)) {
                // `extra` are `AbstractControlOptions`
                validators = extra.validators != null ? extra.validators : null;
                asyncValidators = extra.asyncValidators != null ? extra.asyncValidators : null;
                updateOn = extra.updateOn != null ? extra.updateOn : undefined;
            }
            if (isLegacyOrOpts(extra)) {
                // `extra` are legacy form group options
                validators = validators || [];
                if (extra.validator != null) {
                    validators.push(extra.validator);
                }
                asyncValidators = asyncValidators || [];
                if (extra.asyncValidator != null) {
                    asyncValidators.push(extra.asyncValidator);
                }
            }
        }
        else {
            extra = {};
        }
        if (this.options?.classValidatorOptions && !extra.classValidatorOptions) {
            extra.classValidatorOptions = this.options?.classValidatorOptions;
        }
        if (this.options?.classTransformOptions && !extra.classTransformOptions) {
            extra.classTransformOptions = this.options?.classTransformOptions;
        }
        if (this.options?.classTransformToPlainOptions && !extra.classTransformToPlainOptions) {
            extra.classTransformToPlainOptions = this.options?.classTransformToPlainOptions;
        }
        if (this.options?.validateAllFormFields !== undefined && extra.validateAllFormFields === undefined) {
            extra.validateAllFormFields = this.options?.validateAllFormFields;
        }
        // Set default classValidatorOptions
        if (!extra.classValidatorOptions) {
            extra.classValidatorOptions = DEFAULT_CLASS_VALIDATOR_OPTIONS;
        }
        if (!extra.classTransformOptions) {
            extra.classTransformOptions = DEFAULT_CLASS_TRANSFORM_OPTIONS;
        }
        if (!extra.classTransformToPlainOptions) {
            extra.classTransformToPlainOptions = DEFAULT_CLASS_TRANSFORM_TO_PLAIN_OPTIONS;
        }
        if (extra.validateAllFormFields === undefined) {
            extra.validateAllFormFields = false;
        }
        let newControlsConfig;
        if (controlsConfig !== undefined) {
            newControlsConfig = controlsConfig;
        }
        // experimental
        if (controlsConfig === undefined) {
            newControlsConfig = { ...this.createEmptyObject(factoryModel, { __experimental__: true }) };
            if (newControlsConfig !== undefined) {
                Object.keys(newControlsConfig).forEach((key) => {
                    if (canCreateGroup() && newControlsConfig) {
                        // recursively create a dynamic group for the nested object
                        newControlsConfig[key] = this.group(newControlsConfig[key].constructor, undefined, {
                            classValidatorOptions: extra.classValidatorOptions,
                            classTransformOptions: extra.classTransformOptions,
                            classTransformToPlainOptions: extra.classTransformToPlainOptions,
                            validateAllFormFields: extra.validateAllFormFields,
                            asyncValidators,
                            updateOn,
                            validators,
                        });
                    }
                    else {
                        if (canCreateArray() && newControlsConfig) {
                            if (newControlsConfig[key][0].constructor) {
                                // recursively create an array with a group
                                newControlsConfig[key] = super.array(newControlsConfig[key].map((newControlsConfigItem) => this.group(newControlsConfigItem.constructor, undefined, {
                                    classValidatorOptions: extra.classValidatorOptions,
                                    classTransformOptions: extra.classTransformOptions,
                                    classTransformToPlainOptions: extra.classTransformToPlainOptions,
                                    validateAllFormFields: extra.validateAllFormFields,
                                    asyncValidators,
                                    updateOn,
                                    validators,
                                })));
                            }
                            else {
                                // Create an array of form controls
                                newControlsConfig[key] = super.array(newControlsConfig[key].map((newControlsConfigItem) => this.control(newControlsConfigItem)));
                            }
                        }
                    }
                    function canCreateGroup() {
                        const candidate = newControlsConfig && newControlsConfig[key];
                        return (candidate &&
                            !Array.isArray(candidate) &&
                            candidate.constructor &&
                            typeof candidate === 'object' &&
                            (candidate.length === undefined ||
                                (candidate.length !== undefined && Object.keys(candidate).length === candidate.length)));
                    }
                    function canCreateArray() {
                        if (Array.isArray(newControlsConfig && newControlsConfig[key]) === false) {
                            return false;
                        }
                        const candidate = newControlsConfig && newControlsConfig[key][0];
                        return (candidate.constructor &&
                            typeof candidate === 'object' &&
                            (candidate.length === undefined ||
                                (candidate.length !== undefined && Object.keys(candidate).length === candidate.length)));
                    }
                });
            }
        }
        // Remove empty
        validators = validators && validators.filter((validator) => validator);
        asyncValidators = asyncValidators && asyncValidators.filter((validator) => validator);
        // Create an Angular group from the top-level object
        let classValidators = getClassValidators(factoryModel, newControlsConfig, extra && extra.classValidatorOptions);
        let formGroup = super.group(classValidators, {
            ...(asyncValidators || {}),
            ...(updateOn || {}),
            ...(validators || {}),
        });
        // Initialize the resulting group
        const dynamicFormGroup = this.factoryDynamicFormGroup(factoryModel, newControlsConfig, {
            asyncValidators,
            updateOn,
            validators,
        }, undefined, extra);
        // Add all angular controls to the resulting dynamic group
        Object.keys(formGroup.controls).forEach((key) => {
            dynamicFormGroup.addControl(key, formGroup.controls[key]);
        });
        // Add a listener to the dynamic group for value changes; on change, execute validation
        dynamicFormGroup.subscribeToValueChanges(undefined, extra && extra.classValidatorOptions);
        classValidators = null;
        formGroup = null;
        if (extra.validateAllFormFields) {
            dynamicFormGroup.validate();
        }
        // console.timeEnd(factoryModel.toString());
        return dynamicFormGroup;
    }
    factoryFormBuilder() {
        return new FormBuilder();
    }
    factoryDynamicFormGroup(factoryModel, fields, validatorOrOpts, asyncValidator, options) {
        const formGroup = this.options?.factoryDynamicFormGroup
            ? this.options.factoryDynamicFormGroup(factoryModel, fields, validatorOrOpts, asyncValidator, options)
            : new DynamicFormGroup(factoryModel, fields, validatorOrOpts, asyncValidator, options);
        formGroup.dynamicFormBuilder = this;
        formGroup.originalFormBuilder = this.options?.factoryFormBuilder
            ? this.options?.factoryFormBuilder()
            : this.factoryFormBuilder();
        return formGroup;
    }
    // *******************
    // Helpers
    /**
     * Recursively creates an empty object from the data provided
     */
    createEmptyObject(factoryModel, data = {}) {
        let modifed = false;
        let object = factoryModel ? this.emptyDynamicFormGroup.plainToClass(factoryModel, data) : data;
        let fields = Object.keys(object);
        let objectFieldNameLength;
        let objectFieldName0;
        fields.forEach((fieldName) => {
            const object1 = object[fieldName];
            objectFieldNameLength = object1 && Array.isArray(object1) ? object1.length : undefined;
            if (objectFieldNameLength !== undefined) {
                objectFieldName0 = object1[0];
                if (objectFieldNameLength === 1 && Object.keys(objectFieldName0).length > 0 && objectFieldName0.constructor) {
                    object[fieldName] = [this.createEmptyObject(objectFieldName0.constructor)];
                }
                if (objectFieldNameLength === 0) {
                    data[fieldName] = [{}];
                    modifed = true;
                }
            }
            else {
                data[fieldName] = undefined;
            }
        });
        if (modifed) {
            object = null;
            fields = null;
            return this.createEmptyObject(factoryModel, data);
        }
        fields = null;
        return object;
    }
}

/*
 * Public API Surface of ngx-dynamic-form-builder
 */

/**
 * Generated bundle index. Do not edit.
 */

export { DEFAULT_CLASS_TRANSFORM_OPTIONS, DEFAULT_CLASS_TRANSFORM_TO_PLAIN_OPTIONS, DEFAULT_CLASS_VALIDATOR_OPTIONS, DynamicFormBuilder, DynamicFormControl, DynamicFormGroup, ExposeNested, FOREVER_INVALID_NAME, foreverInvalid, getClassValidators, getGlobal, getOrSetEmptyObject, getValidatorMessagesStorage, getValidatorTitlesStorage, hasToJSON, isAbstractControlOptions, isDynamicFormGroupConfig, isLegacyOrOpts, mergeErrors, transformValidationErrors, updateValidatorMessagesStorage, updateValidatorTitlesStorage };
//# sourceMappingURL=teatimeforever-ngx-dynamic-form-builder.mjs.map
